// Sketch for running FrSky V8 protocol on a Corona transmitter module
// ceptimus November 2020
// uses a cut-down-to-fit-an-ATmega48 code from my FrSkyV8Tx library, which in turn was derived from
// https://github.com/pascallanger/DIY-Multiprotocol-TX-Module by Midelic and Pascal Langer(hpnuts)
//
// Corona transmitter module has ATmega48 8MHz external clock
//
// ISCP programming port on the module is the six-pin 0.1-inch pitch header labelled P5.
// Starting with pin 1 (square pad) 1:MISO 2:VCC 3:SCK 4:MOSI 5:RST 6:GND
// you can solder in some header pins, or just hold in a connector for one-off programming.
// Must use 3.3V USBasp.  Check all USBasp signal pins (not just VCC) are 3.6V max before connecting.
//
// The other 3-pin port, labelled P4, is a serial port
// Pin 1 (square):GND, 2:TX(Arduino pin 1) 3:RX(Arduino pin 0)
//
// The 2-color LED is driven by A0 (Green) and A1 (Red)
// The bind button connects A2 to GND when pressed
//
// The RF CC2500 sub-board pinout - starting on the pin very furthest from the antenna pigtail
// Csn:Arduino pin-10 GD00:pin-2 GD02:pin-3 SCK:pin-13 SO:pin-12 SI:pin-11 (standard SPI-connection)
// On the other side, starting furthest from the antenna pigtail
// Arduino pin-9, VCC, pin-5, pin-7, pin-6, GND - these are power, plus something to do with PA/LNA
// drive pin-9 HIGH, 5 and 6 LOW, 7 HIGH
//
// The PPM signal goes to Arduino Pin 8. via cap and resistor for Futaba, via transistor for JR

#include <EEPROM.h>

// green LED A0, red LED A1
#define GREEN_OFF() asm("cbi 8, 0")
#define GREEN_ON() asm("sbi 8, 0")
#define RED_OFF() asm("cbi 8, 1")
#define RED_ON() asm("sbi 8, 1")

// pin 10 - PORTB bit 2 - CSn line for CC2500 - active low
#define SELECT_CC2500() asm("cbi 5, 2")
#define DESELECT_CC2500() asm("sbi 5, 2")

// bind button A2
#define BIND_PRESSED() !(PINC & _BV(2))

// cppm signal pin 8
#define CPPM_LOW() !(PINB & _BV(0))

// Serial port for debug printing - but without using Serial because that is too big
#define BAUD 115200
#define MYUBRR ((F_CPU/8 + BAUD/2)/BAUD - 1)
#define BUF_SIZE 80
char buf[BUF_SIZE + 1] = {'\0'}; // default all elements to '\0'
char* bPtr = buf;

volatile int16_t channelMicros[8];
volatile uint32_t timer2cycles = 0UL;
volatile uint8_t numChannels = 0;
volatile uint16_t ch0max = 0;
volatile uint16_t ch0min = 3000;
volatile uint8_t bindCounter; // counts 200 down to 1 while binding.  0 when not binding
uint8_t packetBuffers[48];
uint16_t txID;
uint16_t seed;
uint8_t phase;
uint8_t crcID;
uint8_t txPower;
uint8_t oldTxPower;
int8_t fineTune;
int8_t oldFineTune;
bool transmitting;

const PROGMEM uint8_t conf[][2] = { // address:data pairs table for configuring CC2500
  {0x02, 0x06}, // IOCFG0
  {0x00, 0x06}, // IOCFG2
  {0x17, 0x0C}, // MCSM1
  {0x18, 0x18}, // MCSM0
  {0x06, 0xFF}, // PKTLEN
  {0x07, 0x04}, // PKTCTRL1
  {0x08, 0x05}, // PKTCTRL0
  {0x3E, 0xFF}, // PATABLE
  {0x0B, 0x08}, // FSCTRL1
  {0x0C, 0x00}, // FSCTRL0
  {0x0D, 0x5C}, // FREQ2
  {0x0E, 0x58}, // FREQ1
  {0x0F, 0x9D}, // FREQ0
  {0x10, 0xAA}, // MDMCFG4
  {0x11, 0x10}, // MDMCFG3
  {0x12, 0x93}, // MDMCFG2
  {0x13, 0x23}, // MDMCFG1
  {0x14, 0x7A}, // MDMCFG0
  {0x15, 0x41}, // DEVIATN
  {0x19, 0x16}, // FOCCFG
  {0x1A, 0x6C}, // BSCFG
  {0x1B, 0x43}, // AGCCTRL2
  {0x1C, 0x40}, // AGCCTRL1
  {0x1D, 0x91}, // AGCCTRL0
  {0x21, 0x56}, // FREND1
  {0x22, 0x10}, // FREND0
  {0x23, 0xA9}, // FSCAL3
  {0x24, 0x0A}, // FSCAL2
  {0x25, 0x00}, // FSCAL1
  {0x26, 0x11}, // FSCAL0
  {0x29, 0x59}, // FSTEST
  {0x2C, 0x88}, // TEST2
  {0x2D, 0x31}, // TEST1
  {0x2E, 0x0B}, // TEST0
  {0x03, 0x07}, // FIFOTHR
  {0x09, 0x00}, // ADDR
  {0x00, 0x2F}, // IOCFG2 - TX_EN
  {0x02, 0x6F}, // IOCFG0 - TX_EN
  {0xFF, 0xFF}  // end marker
};

void setChannel(uint8_t channelNo, int16_t us) { // channelNo 0 to 7, pulse width to servo in microseconds
  if (channelNo >= 8) {
    return;
  }
  if (us < 732) { // standard travel is 1500 +/- 512us
    us = 732; // limit to absolute maximum 150% of standard travel
  } else if (us > 2268) {
    us = 2268;
  }
  us += us >> 1; // V8 protocol uses 1.5 counts per microsecond
  uint8_t *p = packetBuffers + (channelNo < 4 ? 6 : 16 + 6) + ((channelNo & 0x03) << 1); // pointer to buffer low byte
  uint16_t *p16 = (uint16_t *)p; // pointer to update two bytes more atomically
  *p16 = us;
}

void CC2500strobe(uint8_t state) {
  SELECT_CC2500();
  SPIwrite(state);
  DESELECT_CC2500();
}

extern "C" {
  void bitBangSPIout(volatile uint8_t* mosiPort, uint8_t mosiMask, volatile uint8_t* sckPort, uint8_t sckMask, uint8_t b);
}

void SPIwrite(uint8_t b) {
  bitBangSPIout(&PORTB, _BV(3), &PORTB, _BV(5), b);
}

void CC2500writeReg(uint8_t address, uint8_t b) {
  SELECT_CC2500();
  SPIwrite(address);
  SPIwrite(b);
  DESELECT_CC2500();
}

void CC2500writeRegisterMulti(uint8_t address, uint8_t *b, uint8_t len) {
  SELECT_CC2500();
  SPIwrite(0x40 | address); // WRITE_BURST
  for (uint8_t i = 0; i < len; i++) {
    SPIwrite(*b++);
  }
  DESELECT_CC2500();
}

void CC2500writeData(uint8_t *b, uint8_t len) {
  CC2500strobe(0x3B); // SFTX - flush the TX FIFO buffer
  CC2500writeRegisterMulti(0x3F, b, len); // TX FIFO
  CC2500strobe(0x35); // STX - Enable Tx
}

uint8_t crc(uint8_t b, uint8_t *d, uint8_t len) {
  for (uint8_t i = 0; i < len; i++)
  {
    b ^= *d++;
    for (uint8_t j = 0; j < 8; j++) {
      if (b & 0x80) {
        b = (b << 1) ^ 0x07;
      } else {
        b <<= 1;
      }
    }
  }
  return b;
}

void setTransmitterID(uint16_t transmitterID) {
  packetBuffers[2] = packetBuffers[16 + 2] = (uint8_t)((transmitterID >> 8) & 0x007FU); // limited to 15 bits
  packetBuffers[1] = packetBuffers[16 + 1] = (uint8_t)(transmitterID & 0x00FFU);
  packetBuffers[32 + 4] = packetBuffers[2]; // binding packet has transmitterID at offsets [3][4]
  packetBuffers[32 + 3] = packetBuffers[1]; // instead of [1][2]
  if (transmitting) {
    stopTransmitting();
    startTransmitting();
  }
}

void startTransmitting(void) {
  uint8_t addr = 0;
  uint8_t b = 0;
  uint8_t i = 0;

  if (transmitting) return;

  while ((addr != 0xFF) || (b != 0xFF)) {
    addr = pgm_read_byte_near(&conf[i][0]);
    b = pgm_read_byte_near(&conf[i][1]);
    if (addr == 0x0C) { // FSCTRL0
      oldFineTune = b = fineTune;
    } else if (addr == 0x3E) { // PATABLE
      oldTxPower = b = txPower;
    }
    CC2500writeReg(addr, b);
    i++;
  }
  CC2500strobe(0x36); // SIDLE - set idle state
  seed = 1;
  phase = 0;
  crcID = 0xD6; // compute crc on the transmitterID
  for (i = 0; i < 2; i++) {
    crcID ^= packetBuffers[2 - i];
    for (uint8_t j = 0; j < 8; j++) {
      if (crcID & 0x01) {
        crcID = (crcID >> 1) ^ 0x83;
      } else {
        crcID >>= 1;
      }
    }
  }
  // use timer/counter 1 to generate timer events every 9006us in normal mode
  // or every 53456us in bind mode
  noInterrupts();
  TCCR1A = 0x00;
  TCCR1B = 0x00;
  TCNT1 = 0x0000;
  OCR1A = bindCounter ? (F_CPU <= 8000000 ? 6682 - 1 : 2 * 6682 - 1) : (F_CPU <= 8000000 ? 9006 - 1 : 2 * 9006 - 1);
  TIMSK1 = 0x02; // enable compare match interrupt
  TCCR1B = bindCounter ? 0x0B : 0x0A; // start in CTC mode with prescale /8 for normal use or prescale /64 when binding
  interrupts();
  transmitting = true;
}

void stopTransmitting(void) {
  noInterrupts();
  TCCR1A = 0x00;
  TCCR1B = 0x00;
  interrupts();
  delay(10);
  CC2500strobe(0x36); // SIDLE - set idle state
  bindCounter = 0;
  transmitting = false;
}

// normal micros() only resolves to 8 us on 8MHz Arduinos (or 4 us at 16 MHz)
// this version resolves to 1 us at both clock frequencies
uint32_t accurateMicros(void) {
  uint8_t sreg = SREG;
  cli();
  uint32_t u = timer2cycles;
  uint8_t t = TCNT2;
  if (TIFR2 & 0x01) {
    t = TCNT2;
    timer2cycles = ++u;
    TIFR2 |= 0x01;
  }
  SREG = sreg;
#if F_CPU <= 8000000
  return (u << 8) | t;
#else
  return (u << 7) | (t >> 1);
#endif
}

void setup() {
  DDRB |= 0b00101110; // pins 9, 10, MOSI, and SCK OUTPUT
  DDRC |= 0b00000011; // LEDS on A0 and A1 OUTPUT
  DDRD |= 0b11100000; // pins 5, 6, 7 OUTPUT
  PORTB |= 0b00000010; // pin 9 HIGH
  PORTC |= 0b00000100; // bind button on A2 INPUT_PULLUP
  PORTD &= 0b10011111; // pins 5, 6 LOW
  PORTD |= 0b10000000; // pin 7 HIGH

  DESELECT_CC2500();
  packetBuffers[0] = packetBuffers[16] = packetBuffers[32] = 14; // data length
  packetBuffers[5] = 0x0F; // first four channels packet indicator (swapped to 0x00 for phase 4)
  packetBuffers[16 + 5] = 0xF0; // last four channels packet indicator - always 0xF0
  for (uint8_t ch = 0; ch < 8; ch++) {  // default channels to mid-position (1.5 ms)
    setChannel(ch, 1500);
  }
  setChannel(2, 988); // default channel 3 (probably throttle) to 100% low.
  packetBuffers[32 + 1] = 0x03; // binding packet setup
  packetBuffers[32 + 2] = 0x01;
  packetBuffers[32 + 11] = packetBuffers[32 + 12] = packetBuffers[32 + 13] = 0;
  bindCounter = 0;
  transmitting = false;

  EEPROM.get(0, txID); // retrieve from EEPROM
  txID &= 0x7FFF; // FrSky V protocol IDs are 15-bit
  setTransmitterID(txID);
  EEPROM.get(2, fineTune); // retrieve from EEPROM
  txPower = 255;

  delay(500); // allow start up time
  while (BIND_PRESSED()) { // no LED while they press it
    bindCounter = 200; // but we'll bind once they release it
    // packet[5] is the binding index and cycles 0 5 10 15 20 25 30 35 40 45 0 5 ... while binding
    // packet[6] to packet[10] are derived from packet[5]
    packetBuffers[32 + 5] = 0;
  }
  startTransmitting(); // either in bind mode or normally
  if (millis() > 10000UL) { // they pressed bind for ten seconds +
    stopTransmitting();
    txID = ((micros() >> 3) & 0x7FFFUL); // essentially, a 15-bit random number
    EEPROM.put(0, txID);
    setTransmitterID(txID);
    // now enter a roving bind.  Keep changing the fine tune.
    // eventually the receiver will bind, and then the user will cycle transmitter power to
    // get out of the roving bind mode.
    // flash alternate red/green in roving bind mode.
    while (true) {
      if ((millis() % 1000UL) < 500UL) {
        RED_OFF();
        GREEN_ON();
      } else {
        GREEN_OFF();
        RED_ON();
      }
      if (bindCounter == 0) {
        delay(20);
        stopTransmitting();
        delay(100);
        txPower = bindCounter = 100;
        packetBuffers[32 + 5] = 0;
        fineTune += 16;
        startTransmitting();
      }
    }

  } else if (bindCounter) { // shorter press starts a normal bind for about ten seconds
    while (bindCounter) { // during which time, we'll flash the red LED
      if ((bindCounter % 20) > 9) {
        RED_ON();
      } else {
        RED_OFF();
      }
    }
  }
  RED_ON(); // solid red LED for normal operation

  // cppm signal on pin 8 (port B bit 0)
  PORTB |= _BV(0); // cppm signal on 8 INPUT_PULLUP
  PCMSK0 = 0x01; // only allow port b pin change interrupts for pin 8
  PCICR |= _BV(0); // pin change interrupts on port b

  // Serial port setup - for debug messages
  UBRR0H = (uint8_t)(MYUBRR >> 8);
  UBRR0L = (uint8_t)MYUBRR; // set baud rate
  UCSR0A = 1 << U2X0;
  UCSR0B = 1 << TXEN0; // enable transmitter
  UCSR0C = 3 << UCSZ00; // 8N1 format

  // Accurate microseconds using timer 2
  TCCR2A = 0x00;
  TCCR2B = 0x02; // /8 clock prescale
  TIMSK2 |= 0x01;
}

char* append(char* p, char c) { // serial printing - print a character
  if ((p - buf) < BUF_SIZE) {
    *p++ = c;
  }
  return p;
}

char* append(char* bP, const char *s) { // serial printing - print a string
  while (*s) {
    bP = append(bP, *s++);
  }
  return bP;
}

// serial printing - unsigned int, width > 5 allows right-aligned formatting
// could be improved to allow widths down to one - but if value exceeds width
// then print the value.  Also, correctly formatted negative numbers would be nice.
char* append(char* p, uint16_t v, int width) { 
  while (width > 5) {
    p = append(p, ' ');
    width--;
  }
  bool leadSpace = true;
  for (uint16_t powerTen = 10000; powerTen; powerTen /= 10) {
    char digit = '0';
    while (v >= powerTen) {
      digit++;
      v -= powerTen;
    }
    if (leadSpace && digit == '0' && powerTen > 1) {
      digit = ' ';
    } else {
      leadSpace = false;
    }
    p = append(p, digit);
  }
  return p;
}

void loop() {
  //static uint32_t nextPrintMillis = millis();
  static uint32_t millisBindPressedAt = millis();
  bool buttonState, previousState;

  if (BIND_PRESSED()) {
    // 10-second press of bind in normal mode enters fine tuning mode
    if ((millis() - millisBindPressedAt) > 10000UL) {
      millisBindPressedAt = millis();
      RED_OFF();
      // remain in fine tune till ten seconds without a press have elapsed
      while ((millis() - millisBindPressedAt) < 10000UL) {
        if ((millis() % 500UL) < 250) { // rapid green flashing in fine tune mode
          GREEN_ON();
        } else {
          GREEN_OFF();
        }
        buttonState = BIND_PRESSED();
        // change fine tuning each time they press the button
        if (buttonState && (buttonState != previousState)) {
          millisBindPressedAt = millis();
          fineTune += 8; // change tuning
          GREEN_OFF();
          RED_ON();
          delay(500); // flash red LED to acknowledge eash button press/tuning change
          RED_OFF();
        }
        previousState = buttonState;
      }
      GREEN_OFF(); // back to normal mode
      RED_ON();
      EEPROM.update(2, fineTune); // store current tuning (if changed)
    }
  } else {
    millisBindPressedAt = millis();
  }

  if (UCSR0A & (1 << UDRE0)) { // serial port ready for next character
    if (*bPtr) {
      UDR0 = *bPtr; // send a character
      *bPtr++ = '\0';
    } else {  // buf sent - prepare next message.  Typical example for 8-channel CPPM signal
              // 8: 1541 1545 1417 1552 1014 1718 1544 1532 ( 1538: 1546) 15448 -    9
              // 8: is number of channels, then current channel readings
              // numbers in parentheses are min and max readings for 1st channel
              // 15448 is the transmitter ID, -9 is the fine tuning
      bPtr = append(buf, numChannels, 1);
      bPtr = append(bPtr, ':');
      for (uint8_t i = 0; i < numChannels; i++) {
        bPtr = append(bPtr, channelMicros[i], 4);
      }
      bPtr = append(bPtr, " (");
      bPtr = append(bPtr, ch0min, 5);
      bPtr = append(bPtr, ':');
      bPtr = append(bPtr, ch0max, 5);
      bPtr = append(bPtr, ')');
      bPtr = append(bPtr, txID, 6);
      bPtr = append(bPtr, ' ');
      if (fineTune < 0) {
        bPtr = append(bPtr, '-');
        bPtr = append(bPtr, -fineTune, 5);
      } else {
        bPtr = append(bPtr, fineTune, 5);        
      }
      bPtr = append(bPtr, '\n');
      bPtr = buf;
    }
  }
}

ISR(PCINT0_vect) { // pin change interrupt from CPPM signal Arduino pin 8 (port B, bit 0)
  static uint32_t previousMicros = accurateMicros();
  static uint8_t channel = 0;
  static uint8_t notInSync = 2; // wait for 2nd valid frame before accepting values

  uint32_t now = accurateMicros();
  if (CPPM_LOW()) { // falling edge
    sei();
    uint32_t interval = now - previousMicros;
    previousMicros = now;
    if (interval > 3000) {
      channel = 0;
      if (notInSync) --notInSync; // at startup, count first two sync pulses before accepting values
    } else if (!notInSync && channel < 8) {
      setChannel(channel, interval);
      if (!channel) {
        if (interval > ch0max) {
          ch0max = interval;
        }
        if (interval < ch0min) {
          ch0min = interval;
        }
      }
      channelMicros[channel++] = interval;
      if (channel > numChannels) {
        numChannels = channel;
      }
    }
  }
}

ISR(TIMER1_COMPA_vect) { // timer1 interrupt every 9006us in normal mode or every 53456us in bind mode
  sei(); // allow other interrupts while servicing this timer interrupt
  int packetStart = 0;
  if (bindCounter) { // if binding
    if (bindCounter == 200) { // reduce transmitter power while binding
      oldTxPower = 100; // normal power will be resumed after binding
      CC2500writeReg(0x3E, oldTxPower); // PATABLE
    }
    packetBuffers[32 + 6] = packetBuffers[32 + 5] * 5 + 6;
    packetBuffers[32 + 7] = packetBuffers[32 + 5] * 5 + 11;
    packetBuffers[32 + 8] = packetBuffers[32 + 5] * 5 + 16;
    packetBuffers[32 + 9] = packetBuffers[32 + 5] * 5 + 21;
    packetBuffers[32 + 10] = packetBuffers[32 + 5] * 5 + 26;
    CC2500strobe(0x36); // SIDLE - set idle state
    CC2500writeReg(0x0A, 0x00); // channel number 0 while binding
    packetBuffers[32 + 14] = crc(0x93, packetBuffers + 32, 14);
    CC2500writeData(packetBuffers + 32, 15);
    packetBuffers[32 + 5] += 5;
    if (packetBuffers[32 + 5] > 45) {
      packetBuffers[32 + 5] = 0;
    }
    if (--bindCounter == 0) { // finished binding - back to normal operation
      OCR1A = F_CPU <= 8000000 ? 9006 - 1 : 2 * 9006 - 1;
      TCCR1B = 0x0A; // CTC mode with prescale /8 when not binding
    }
  } else { // not binding
    seed = (uint16_t)((0xAAUL * (uint32_t)seed) % 0x7673);
    // channel = (uint8_t)((seed & 0x00FF) % 0x32);
    CC2500strobe(0x36); // SIDLE - set idle state
    if (fineTune != oldFineTune) {
      CC2500writeReg(0x0C, fineTune); // FSCTRL0
      oldFineTune = fineTune;
    }
    CC2500writeReg(0x0A, (uint8_t)(((uint8_t)((seed & 0x00FF) % 0x32)) * 5 + 6)); // channel number
    packetStart = (phase & 0x01) ? 16 : 0; // send first four channels on even phases, and last four on odd
    packetBuffers[packetStart + 3] = (uint8_t)(seed & 0x00FF);
    packetBuffers[packetStart + 4] = (uint8_t)((seed >> 8) & 0x00FF);
    if (phase == 0) {
      packetBuffers[5] = 0x0F; packetBuffers[5] = 0x0F;
    } else if (phase == 4) {
      packetBuffers[5] = 0x00;
    }
    packetBuffers[packetStart + 14] = crc(crcID, packetBuffers + packetStart, 14);
    if (++phase == 5) {
      if (txPower != oldTxPower) {
        CC2500writeReg(0x3E, txPower); // PATABLE
        oldTxPower = txPower;
      }
      phase = 0;
    }
    CC2500writeData(packetBuffers + packetStart, 15);
  }
}

ISR(TIMER2_OVF_vect) {
  ++timer2cycles;
}
